package com.winhc.phoenix.example.service.impl;

import com.winhc.phoenix.example.dao.SearchDao;
import lombok.AllArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.elasticsearch.common.lucene.search.function.CombineFunction;
import org.elasticsearch.index.query.BoolQueryBuilder;
import org.elasticsearch.index.query.MultiMatchQueryBuilder;
import org.elasticsearch.index.query.Operator;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.index.query.functionscore.FunctionScoreQueryBuilder;
import org.elasticsearch.index.query.functionscore.ScriptScoreFunctionBuilder;
import org.elasticsearch.script.Script;
import org.elasticsearch.script.ScriptType;
import org.elasticsearch.search.rescore.QueryRescoreMode;
import org.elasticsearch.search.rescore.QueryRescorerBuilder;
import org.elasticsearch.search.sort.ScriptSortBuilder;
import org.elasticsearch.search.sort.SortOrder;
import org.springframework.stereotype.Service;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static org.elasticsearch.index.query.QueryBuilders.*;

/**
 * @author: XuJiakai
 * 2021/7/28 14:38
 */
@Slf4j
@Service
@AllArgsConstructor
public class FindRelationshipService {
    public static enum FindType {
        NAME, LEGAL_REPRESENTATIVE, HOLDER_OR_STAFF, APP
    }

    private SearchDao searchDao;

    public static final String index = "winhc-company-v8_3";
    public static final String type = "company";

    public Object findRelationship(String content, List<FindType> typeList, Integer from, Integer size) {
        BoolQueryBuilder boolQuery = getBoolQuery(content, typeList);

        Map<String, Object> map = new HashMap<String, Object>() {{
            put("query_content", content);
        }};
        ScriptSortBuilder scriptSortBuilder = new ScriptSortBuilder(new Script(ScriptType.STORED, null, "company-search-script", map), ScriptSortBuilder.ScriptSortType.NUMBER).order(SortOrder.DESC);

//        QueryRescorerBuilder rescorerBuilder = new QueryRescorerBuilder(functionScoreQuery(new ScriptScoreFunctionBuilder(scriptSortBuilder.script())))
//                .windowSize(50)
//                .setScoreMode(QueryRescoreMode.Multiply);


        QueryRescorerBuilder rescorerBuilder2 = new QueryRescorerBuilder(getRescorerBool(content, typeList))
                .windowSize(50)
                .setScoreMode(QueryRescoreMode.Total);

        Object search = searchDao.search(index, type, boolQuery, Arrays.asList(rescorerBuilder2), null, null, from, size);

        return search;
    }

    private static FunctionScoreQueryBuilder getRescorerBool(String content, List<FindType> typeList) {
        BoolQueryBuilder boolQuery = QueryBuilders.boolQuery();

        if (typeList.contains(FindType.NAME)) {
            boolQuery.should(termQuery("cname.value.keyword", content).boost(1000));
            boolQuery.should(
                    disMaxQuery()
                            .add(multiMatchQuery(content)
                                    .type(MultiMatchQueryBuilder.Type.CROSS_FIELDS)
                                    .minimumShouldMatch("5<90%")
                                    .tieBreaker(0.3F)

                                    .field("cname.value", 16)
                                    .field("history_name.value", 12))
                            .add(multiMatchQuery(content)
                                    .operator(Operator.AND)
                                    .type(MultiMatchQueryBuilder.Type.CROSS_FIELDS)
                                    .tieBreaker(0.3F)
                                    .field("cname.value.standard", 16)
                                    .field("history_name.value.standard", 12))
                            .tieBreaker(0.4F)
            );

            boolQuery.should(disMaxQuery()
                    .add(disMaxQuery()
                            .add(termQuery("holder.name.keyword", content).boost(5.5F))
                            .tieBreaker(0.3F)
                    ).add(disMaxQuery()
                            .add(matchPhraseQuery("holder.name", content).boost(10).slop(3))
                            .tieBreaker(0.3F)
                    ).tieBreaker(0.3F)
            );
            boolQuery.should(disMaxQuery()
                    .add(disMaxQuery()
                            .add(termQuery("app_info.keyword", content).boost(40))
                            .add(termQuery("company_tm.keyword", content).boost(20))
                            .tieBreaker(0.4F))
                    .add(disMaxQuery()
                            .add(matchPhraseQuery("app_info", content).boost(19).slop(3))
                            .add(matchPhraseQuery("company_tm", content).boost(7).slop(3))
                            .tieBreaker(0.3F)
                    ).tieBreaker(0.4F)
            );
        }

        if (typeList.contains(FindType.HOLDER_OR_STAFF)) {
            boolQuery.should(disMaxQuery()
                    .add(disMaxQuery()
                            .add(termQuery("holder.name.keyword", content).boost(5.5F))
                            .add(termQuery("staff.name.keyword", content).boost(5.5F))
                            .tieBreaker(0.3F)
                    ).add(disMaxQuery()
                            .add(matchPhraseQuery("holder.name", content).boost(10).slop(3))
                            .add(matchPhraseQuery("staff.name", content).boost(6).slop(3))
                            .tieBreaker(0.3F)
                    ).tieBreaker(0.3F)
            );
        }

        if (typeList.contains(FindType.APP)) {
            boolQuery.should(disMaxQuery()
                    .add(termQuery("app_info.keyword", content).boost(40))
                    .add(matchPhraseQuery("app_info", content).boost(19).slop(3))
                    .tieBreaker(0.4F)
            );
        }


        if (typeList.contains(FindType.LEGAL_REPRESENTATIVE)) {
            BoolQueryBuilder legalRepresentativeQuery = QueryBuilders.boolQuery();
            legalRepresentativeQuery.should(termQuery("legal_entity_name.keyword", content).boost(10));

            legalRepresentativeQuery.should(matchQuery("legal_entity_name", content).boost(6).minimumShouldMatch("5<95%"));
            boolQuery.should(legalRepresentativeQuery);
        }


//        return boolQuery;

        Map<String, Object> map = new HashMap<String, Object>() {{
            put("query_content", content);
        }};
        ScriptSortBuilder scriptSortBuilder = new ScriptSortBuilder(new Script(ScriptType.STORED, null, "company-search-script", map), ScriptSortBuilder.ScriptSortType.NUMBER).order(SortOrder.DESC);

        FunctionScoreQueryBuilder function = functionScoreQuery(boolQuery, new ScriptScoreFunctionBuilder(scriptSortBuilder.script()))
                .maxBoost(100)
                .boostMode(CombineFunction.MULTIPLY);

        return function;
    }


    private BoolQueryBuilder getBoolQuery(String content, List<FindType> typeList) {
        BoolQueryBuilder boolQuery = boolQuery();
        if (typeList.contains(FindType.NAME)) {
            BoolQueryBuilder nameQuery = QueryBuilders.boolQuery();
            nameQuery.should(termQuery("cname.value.keyword", content).boost(1000));
            nameQuery.should(
                    disMaxQuery()
                            .add(multiMatchQuery(content)
                                    .type(MultiMatchQueryBuilder.Type.CROSS_FIELDS)
                                    .minimumShouldMatch("5<90%")
                                    .tieBreaker(0.3F)

                                    .field("cname.value", 16)
                                    .field("history_name.value", 12))
                            .add(multiMatchQuery(content)
                                    .operator(Operator.AND)
                                    .type(MultiMatchQueryBuilder.Type.CROSS_FIELDS)
                                    .tieBreaker(0.3F)
                                    .field("cname.value.standard", 16)
                                    .field("history_name.value.standard", 12))
                            .tieBreaker(0.4F)
            );
            boolQuery.should(nameQuery);
        }

        if (typeList.contains(FindType.LEGAL_REPRESENTATIVE)) {
            BoolQueryBuilder legalRepresentativeQuery = QueryBuilders.boolQuery();
            legalRepresentativeQuery.should(termQuery("legal_entity_name.keyword", content).boost(10));

            legalRepresentativeQuery.should(
                    matchQuery("legal_entity_name", content).boost(6).minimumShouldMatch("5<95%")
            );
            boolQuery.should(legalRepresentativeQuery);
        }

        if (typeList.contains(FindType.HOLDER_OR_STAFF)) {
            boolQuery.should(disMaxQuery()
                    .add(disMaxQuery()
                            .add(termQuery("holder.name.keyword", content).boost(5.5F))
                            .add(termQuery("staff.name.keyword", content).boost(5.5F))
                            .tieBreaker(0.3F)
                    ).add(disMaxQuery()
                            .add(matchPhraseQuery("holder.name", content).boost(10).slop(3))
                            .add(matchPhraseQuery("staff.name", content).boost(6).slop(3))
                            .tieBreaker(0.3F)
                    ).tieBreaker(0.3F)
            );
        }

        if (typeList.contains(FindType.APP)) {
            boolQuery.should(disMaxQuery()
                    .add(termQuery("app_info.keyword", content).boost(40))
                    .add(matchPhraseQuery("app_info", content).boost(19).slop(3))
                    .tieBreaker(0.4F)
            );
        }

        return boolQuery()
                .filter(termQuery("deleted", "0"))
                .filter(rangeQuery("company_score_weight").gt(0.3F))
                .must(boolQuery);
    }
}
